/** \file ntpwatch.c 
    \brief a single-file body of ntpwatch

 */


/* The speed of clock is controlled by tick and freq. 1 tick
   changes the speed of about 1/10000 times changing 1 freq unit
   changes by 2**-36 times. */
#define TICK_MULT 1e4

/* This is 2**36 */
#define FREQ_MULT 68719476736.0
#define NTP_SOCKET 10

/* This must be enough for setting STA_SYNC */
#define SYNC_THRESHOLD 0.1

/* 
   Minimum wait time between non-broadcast NTP packets 
   This should never be really used.
 */
#define NTP_MINIMUM_WAIT 3.0




/* Define DPTR_LEVEL to 0 to get uncompiled dptr.
   0 -- no messages, but still working asserts unless NDEBUG is defined
   1 -- essential warnings (skipped NTP)
   3 -- procedure announces
   4 -- parameters and goals
   9 -- everything
   
 */
#ifndef DPTR_LEVEL
#define DPTR_LEVEL 0
#endif
#define JAN_1970        0x83aa7e80      /* 2208988800 -- 1970-1900 in seconds */
/** \brief Convert 64-bit NTP field into real value in seconds since 1970, i.e. UNIX time 
 */
#define rtime(field)                                                  \
  (ntohl(pkt.ntp.field.s) - JAN_1970 +				      \
   ntohl(pkt.ntp.field.ns)*(1.0/4294967296.0) + ntp_delay_mean)

#define rtimediff(f1,f2)                                              \
  ( (((signed long)ntohl(pkt.ntp.f1.s))-((signed long)ntohl(pkt.ntp.f2.s)))     \
    +(((signed long)ntohl(pkt.ntp.f1.ns))-((signed long)htonl(pkt.ntp.f2.ns)))*(1.0/4294967296.0))


/* difference between to timeval.
   Returns t1-t0
   It assumes tv_sec and tv_usec are signed
 */
#define timevaldiff(t1,t0) ((((signed long)t1.tv_usec) - (signed long)t0.tv_usec)*1.0e-6 \
			    + ((signed long)t1.tv_sec-(signed long)t0.tv_sec))



#include "timex.h"
#include <time.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <assert.h>




#ifdef NDEBUG
#undef DPTR_LEVEL
#define DPTR_LEVEL 0
#endif

#if DPTR_LEVEL > 0
#include <stdio.h>
/** \brief sends a debug message. On non-debug system does nothing */
#define dpr(level, ...)                                     \
  do { if(level<=DPTR_LEVEL) {                              \
       struct timex txx;                                    \
       txx.modes=0; adjtimex(&txx);                         \
       fprintf(stderr,"DEBUG %7.2f %s:%d: ",                \
           timevaldiff(txx.time,prog_start.time),           \
            __FILE__, __LINE__);                            \
       fprintf(stderr, __VA_ARGS__);		            \
       fprintf(stderr,"\n");                                \
    }} while(0)
#else
#define dpr(level, format, ...) assert(1)
#endif





#if 0
#include "conf.h"
#endif

/* Long doubles are faster, albeit double certainly are good
   enough. I tried to keep maximum bits during subtraction, and I
   believe even 4-byte float is good enough */

#ifdef HAVE_LONG_DOUBLE
typedef long double time_real;
typedef long double high_precision;
#else
typedef double high_precision;
typedef double time_real;
#endif


struct ntp_time64 {
  uint32_t s;
  uint32_t ns;
} __attribute__((packed));

struct ntp_packet {
  uint8_t  li_vn_mode;
  uint8_t  stratum, poll;
  int8_t   precision;
  uint32_t root_delay, root_dispersion, reference_id;
  struct ntp_time64 ref, orig, recv, transm;
} __attribute__((packed));

struct time_point 
{
  struct timex host_time;
  time_real server_minus_host; 
};
static struct time_point pts[3];
static struct timex prog_start, now, ntp_last_request;

/* I know tick is unsigned but sometimes it is substracted from
   10000 and it is simply easier to treat it like that */

/* tan0 is the optimal tick and freq. */
high_precision tan0=0;

static void restore_optimal_clock(int signum)
{
  struct timex r;
  r.modes = ADJ_TICK | ADJ_FREQUENCY;
  r.tick = floor(tan0*TICK_MULT+0.5);
  r.freq = (tan0-r.tick*(1.0/TICK_MULT))*FREQ_MULT;
  r.tick += TICK_MULT;
  adjtimex(&r);

  if(signum>0) signal(signum, restore_optimal_clock);
  if(signum!=SIGALRM && signum!=SIGUSR1) 
    _exit(0);
} 


static struct sockaddr_in ntp_server;


/* Generic ABS, be warned about its macroness, argument is evaluated twice */
#define ABS(x) ( (x>=0) ? (x) : (-(x)) )
#define SIG(x) ( (x>=0) ? 1 : -1 )



static uint32_t rnd_seed;
static inline uint32_t random_32bit(void) 
{
  rnd_seed = 1664525 * rnd_seed + 1013904223;
  return rnd_seed;
}

/* Update NTP delay time. Calculate the mean and dispersion */
#define NTP_DELAY_PROBES 8
static time_real ntp_delay_mean, ntp_delay_variance;
static time_real ntp_delay_probes[NTP_DELAY_PROBES];
static unsigned long  ntp_next_probe_idx, ntp_next_signal_time;
static time_real ntp_signal_delay;

static void send_warning(const char *s)
{
  static unsigned long last_warning=0;
  static unsigned long last_delay=0;
  static unsigned long last_warning_burst=0;
  assert(now.modes==0);
  adjtimex(&now);
  if( now.time.tv_sec >= last_warning + last_delay ) {
    last_warning=now.time.tv_sec;
    last_delay += last_delay + 1;
    last_warning_burst=5;
  }
  if(last_warning_burst) {
    last_warning_burst--;
    if(fork()==0) {
      char * const argv[3] = {
	(char *)"./ntpwatch-sendwarning", 
	(char *)s, 
	0};
      execve(argv[0],argv,argv+2);
      _exit(1);
    }
    else {
      int i; 
      wait(&i);
    }
  }
}


#define sqr(x) ((x)*(x))
static void adjust_ntp_delay(time_real delay)
{
  int i;
  time_real sum, sumsq;
  if(ntp_next_probe_idx==NTP_DELAY_PROBES) ntp_next_probe_idx=0;
  assert(ntp_next_probe_idx<NTP_DELAY_PROBES);
  ntp_delay_probes[ntp_next_probe_idx++] = delay;
  for(sum=sumsq=0.0, i=0; i<NTP_DELAY_PROBES; i++) 
    sum   += ntp_delay_probes[i],
    sumsq += sqr(ntp_delay_probes[i]);
  sum /= NTP_DELAY_PROBES;
  ntp_delay_mean = sum;
  ntp_delay_variance = sumsq/NTP_DELAY_PROBES - sqr(sum);
  ntp_signal_delay = 1.19680404*ntp_signal_delay 
    /* Add up to 0.5s */
    + random_32bit()*(1.0/(65536.0*65536.0*2));
  if(ntp_signal_delay>2000) ntp_signal_delay /= 2;
  dpr(9,"mean=%.3g var=%.3g", ntp_delay_mean, sqrt(ntp_delay_variance));
}



union {
  struct ntp_packet ntp;
  struct {
    uint32_t cmd;
    uint32_t address;
  } cmd;
  struct {
    uint32_t version;
    uint32_t address; /* NTP address */
    uint32_t diff; /* Difference in microseconds, odd if negative */
  } info;
  unsigned char buff[300];
} pkt;

static int read_packet(void)
{
  struct sockaddr_in sa;
  int r;
  size_t l;
  l=sizeof(sa);
  r=recvfrom(NTP_SOCKET, &pkt, sizeof(pkt), 0,
             (struct sockaddr *)&sa, &l);
  if(r==-1) 
    if(errno != EINTR || errno != EAGAIN) {
      send_warning("Invalid packet");
      exit(1);
    }
  if(pkt.buff[0]==0xFF) {
    /* special packets */
    if(ntohl(pkt.cmd.cmd)==0xFF000001) {
      if(sa.sin_addr.s_addr != htonl(INADDR_LOOPBACK))
        return -1;
      if(pkt.cmd.address && ntohs(sa.sin_port)>=1024)
        ntp_server.sin_addr.s_addr=pkt.cmd.address;
      pkt.info.version = htonl(1);
      pkt.info.address = ntp_server.sin_addr.s_addr;
      { time_real t = pts[2].server_minus_host;      
        if(t<-1000) t=-1000;
        if(t>1000) t=1000;
        pkt.info.diff = 
          (((uint32_t)floor(1e6*ABS(t))) & 0xFFFFFFFEu)
          | (t>=0?0:1);
      }
      sendto(NTP_SOCKET,&pkt.info,sizeof(pkt.info),0,
             (struct sockaddr *)&sa, l);
    }
    return -1;
  }
  dpr(9, "Just read %d bytes",r);
  if(r!=48) return -1;
  if(ntp_server.sin_addr.s_addr==0)  {
    ntp_server.sin_addr.s_addr=sa.sin_addr.s_addr;
#define _CHR(N) 	(int)(((unsigned char *)&sa.sin_addr.s_addr)[N])
    dpr(9,"Setting server address to %d.%d.%d.%d",
	_CHR(0), _CHR(1), _CHR(2), _CHR(3));
#undef _CHR
  }
  if(sa.sin_addr.s_addr != ntp_server.sin_addr.s_addr)
    return -1;
  return 0;
}


static int read_ntp(time_real timeout)
{
  struct pollfd pfd;
  int r;


  pfd.fd      = NTP_SOCKET;
  pfd.events  = POLLIN;
  pfd.revents = 0;
  errno = 0;

  /*
    Check whether there's a pending packet, if so, ignore
    it. There's is no way to tell how long it waits in the queue.
  */
  adjtimex(&now);
  if( poll(&pfd,1,0) != 0) {
    r = read_packet();
    return -1;
  }

  assert(pfd.fd      == NTP_SOCKET);
  assert(pfd.events  == POLLIN);
  assert(pfd.revents == 0);

  if( ntp_next_signal_time < ntp_last_request.time.tv_sec + NTP_MINIMUM_WAIT ) {
    ntp_next_signal_time = ntp_last_request.time.tv_sec + NTP_MINIMUM_WAIT;
    timeout = NTP_MINIMUM_WAIT;
  }
  
  if(timeout > ntp_next_signal_time-now.time.tv_sec) 
    timeout  = ntp_next_signal_time-now.time.tv_sec;
  if(timeout<1.0) timeout=1.0;

  dpr(9, "ntp_next_signal_time=%lu now=%lu.%06lu", 
      ntp_next_signal_time, 
      (unsigned long)now.time.tv_sec, (unsigned long)now.time.tv_usec);
  dpr(9, "now-next=%lu", now.time.tv_sec-ntp_next_signal_time);
  if(now.time.tv_sec>ntp_next_signal_time && ntp_server.sin_addr.s_addr) {
    ntp_next_signal_time  = now.time.tv_sec + (unsigned)ntp_signal_delay;
    dpr(9,"Next NTP signal in %ld sec (del=%g)\n", ntp_next_signal_time-now.time.tv_sec, ntp_signal_delay);
    /* Prepare and send a packet with NTP request */
    memset(&pkt,0,48);
    pkt.ntp.li_vn_mode = 
      (0<<6) + /* LI */
      (4<<3) + /* VN */
      (3<<0);  /* Mode */
    if(0) pkt.ntp.stratum         = 0;
    if(0) pkt.ntp.poll            = 0;
    pkt.ntp.precision       = -15;
    if(0) pkt.ntp.root_delay      = 0;
    if(0) pkt.ntp.root_dispersion = 0;
    if(0) pkt.ntp.reference_id    = 0;
    errno = 0;
    ntp_last_request.modes=0;
    r = sendto(NTP_SOCKET, &pkt, 48, 0, 
               (struct sockaddr*) &ntp_server, sizeof(ntp_server));
    adjtimex(&ntp_last_request);
    if(0) pkt.ntp.orig.s = htonl(((uint32_t)now.time.tv_sec + JAN_1970));
    if(0) pkt.ntp.orig.ns = htonl((uint32_t)(now.time.tv_usec*4294.967296));
    if(r!=48 && errno!=EINTR) 
      send_warning("Cannot send NTP packet");
    else dpr(9, "Sent NTP packet%s.","");
    timeout=2;
  }


  now.modes = 0;
  dpr(9, "Timeout=%g\n", (double)timeout);
  r=poll(&pfd,1,(unsigned)(timeout*1000.0)+random_32bit()%400); 
  adjtimex(&now);

  /* Do nothing if nothing came */
  if(r==0) {
    dpr(9, "Timeout... %s","");
    return -1;
  }

  r = read_packet();
  if(r) return -1;


  /* Ignore every NTP except broadcast or time message */
  if(pkt.ntp.li_vn_mode != 4 + (4<<3) &&
     pkt.ntp.li_vn_mode != 5 + (4<<3))
    return -1;
  
  /* Recalcute sigmas if this is a single packet */
  if( pkt.ntp.li_vn_mode == 4 + (4<<3) ) {
    dpr(9, "orig:  %-.5f", (double)rtime(orig));
    dpr(9, "recv:  %-.5f", (double)rtime(recv));
    dpr(9, "trasm: %-.5f", (double)rtime(transm));
    dpr(9, "now:   %-.5f", now.time.tv_sec+ 1e-6*now.time.tv_usec );
    assert(now.time.tv_sec>=ntp_last_request.time.tv_sec);
    assert(rtime(transm) >= rtime(recv));
    adjust_ntp_delay
      ( (timevaldiff(now.time,ntp_last_request.time) -
         rtimediff(transm,recv))*0.5);
  }
  else {
    dpr(9, "Received **BROADCAST** NTP%s.","");
    dpr(9, "orig:  %-.5f", (double)rtime(orig));
    dpr(9, "recv:  %-.5f", (double)rtime(recv));
    dpr(9, "trasm: %-.5f", (double)rtime(transm));
    dpr(9, "now:   %-.5f", now.time.tv_sec+ 1e-6*now.time.tv_usec );
  }
  
  pts[2].host_time = now;
  {
    /* This MUST be double precision, even if internally it's kept as float */
    high_precision t;
    assert( sizeof(t)>= 8 );
    t = (unsigned long) ntohl(pkt.ntp.transm.s);
    t -= JAN_1970;
    t -= now.time.tv_sec;
    t += (1.0/4294967296.0)*((unsigned long)ntohl(pkt.ntp.transm.ns));
    t -= now.time.tv_usec*1e-6;
    t += ntp_delay_mean;
    pts[2].server_minus_host = t;
    dpr(9, "time diff=%-.6f", (double)pts[2].server_minus_host);
    t*=t;
    if(t<0.03) {
      now.modes = ADJ_STATUS  | ADJ_ESTERROR | ADJ_MAXERROR;
      now.status &= ~STA_UNSYNC;
      now.esterror = floor(1e6 * sqrt(t + ntp_delay_variance));
      now.maxerror = floor(1e6 * sqrt(t + 3*ntp_delay_variance));
    }
    else if( (now.status & STA_UNSYNC) == 0) {
      now.modes  = ADJ_STATUS;
      now.status |= STA_UNSYNC;
    }
    adjtimex(&now);
    now.modes=0;
  }
  return 0;
}





static void initialize_ntpwatch(void)
{
  int i; 
  /* Initialize ntp_delay to (unreal) value 0 */
  for(i=0; i<NTP_DELAY_PROBES; i++) ntp_delay_probes[i]=0.0;
  ntp_delay_mean=ntp_delay_variance=0.0;
  ntp_next_probe_idx=0;
  ntp_last_request.time.tv_sec=0;
  ntp_signal_delay = ntp_next_signal_time = 5; /* few seconds, 10 is equally fine */
  now.modes=0;
  adjtimex(&now);
  memset(pts,0,sizeof(pts));
  for(i=0; i<30 && pts[0].host_time.time.tv_sec==0; i++) {
    pts[2].host_time.time.tv_sec=0;
    read_ntp(5);
    if(pts[2].host_time.time.tv_sec) {
      pts[0]=pts[1];
      pts[1]=pts[2];
    }
  }
  if(pts[0].host_time.time.tv_sec==0) {
    send_warning("Cannot initialize time, waiting 5 minutes");
    sleep(5*60);
    initialize_ntpwatch();
  }
}




static void fix_time(void)
{
  for(;;) {
    struct timex req;
    long goal;
    
    /* 
       Current tangent of time.
       speed of differences of time divided by time interval
       negative if we're too slow
       positive if we're too fast.
   
    */
    high_precision tan;

    if(  ABS( pts[0].host_time.tick - pts[1].host_time.tick ) > 1 ) 
      ntp_next_signal_time = now.time.tv_sec + (ntp_signal_delay=66);
    
    tan = (pts[0].server_minus_host-pts[1].server_minus_host)
      / timevaldiff(pts[1].host_time.time, pts[0].host_time.time);
    
    /* Current theoretical tangent o ideally synchronized clocks
       i.e. the tangent at which server and host clocks run equaspeedy.
       This the value of tan that should be set to get equaspeed
     */
    tan0 = (pts[1].host_time.tick-TICK_MULT)*(1.0/TICK_MULT)
      + pts[1].host_time.freq*(1.0/FREQ_MULT) - tan;
    dpr(8, "tan=%.5g tan0=%.5g diff=%.5g", tan, tan0, (double)pts[1].server_minus_host);

    /* Calculate the new optimal tick and frequency */
    if(ABS(tan0) >  0.03) {
      /* Something with your clock went really wrong. Don't panic
	 until it persists more than several times */
#if 0
      /* Disabled code. Probably it won't do anything useful */
      if(error_count--<0) send_warning("Invalid clock speed, more than 3% off the ideal");
#endif
      tan0=0;
    }

    req=now;
    req.modes = ADJ_FREQUENCY | ADJ_STATUS;

    assert(ntp_signal_delay<3600);
    goal = floor(ntp_signal_delay);
    if(goal<66) goal=66;
    
    tan = tan0 +pts[1].server_minus_host/goal;
    if( ABS(pts[1].server_minus_host) > 0.1 ) {
      /* Set the clock unsyncd */
      req.status  |= STA_UNSYNC;
    }
    else {
      req.status &= ~STA_UNSYNC;
      req.esterror = 1e6*sqrt( sqr(pts[1].server_minus_host) + ntp_delay_variance);
      req.maxerror = 1e6*sqrt( sqr(pts[1].server_minus_host) + 9*ntp_delay_variance);
      req.modes |= ADJ_ESTERROR | ADJ_MAXERROR;
      if(req.esterror>200000 || req.maxerror>300000) {
        req.modes ^= ADJ_ESTERROR | ADJ_MAXERROR | ADJ_STATUS;
        req.status |= STA_UNSYNC;
      }
    }

    /* If the difference is large, use the slow approach but set the clock as synced.
       Otherwise mark it as unsynched and use the faster approach */

    if(ABS(tan)>0.099) {
      tan=0.099*SIG(tan);

      if(ABS(pts[1].server_minus_host)>3600) {
	if(ABS(pts[1].server_minus_host)>3600*10) 
	  send_warning("Clock is unsynchronized many DAYS! Fix it manually");
	else
	  send_warning("Clock is unsynchronized many HOURS! Fix it manually");
	goal=100000000; /* Fixed in the next step anyway */
      }
      else goal = ABS(pts[1].server_minus_host)*9;
      if(goal>15*60) goal=15*60;
    }

    /* Fix the clock */
    req.tick = floor(tan * TICK_MULT+0.5) + TICK_MULT;
    if(  ABS(req.tick - now.tick)<2 ) 
      req.tick = now.tick;
    else 
      req.modes |= ADJ_TICK;
    
    {
      high_precision t = (tan - (req.tick-TICK_MULT)*(1.0/TICK_MULT));
      assert(ABS(t)<3/TICK_MULT);
      req.freq = t*FREQ_MULT;
      dpr(6, "Setting tick=%ld, frequency=%ld (%.5g), goal=%.5g", req.tick, req.freq, 
          (double)t, (double)goal);
    }
    adjtimex(&req);

    /* Now wait till:
       * 0s -- 15s
            do really nothing, ignore everything, even if server_minus_host>1year
       * 15s -- (goal-100s) 
             do nothing except if thing went really wrong. Really wrong
             means ABS(server_minus_host)>(last_one+average_variance+0.05)

       * goal-100s -- goal
            catch the first NTP signal and use it to tune the clock

       * goal -- goal+5s, 
            start sending ntp requests every second

       * > goal+5s
            assume the reference clock is not responding, do an emergency 

       In all cases, if now-pts[2].host_time < 3s update clock errors

    */

    goal += req.time.tv_sec;
    for(;;) {
      int wait;
      static int ntp_error_count=0;
      dpr(9,"now.sec-req.sec=%d, goal=%ld", 
          now.time.tv_sec - req.time.tv_sec, 
          goal); 

      if( now.time.tv_sec < req.time.tv_sec ) {
	send_warning("Emergency. Time flows backward!");
	return;
      }

      wait = goal - now.time.tv_sec;

      if(wait<-5) {
        send_warning("Cannot get NTP packet. Waiting 5 minutes");
        restore_optimal_clock(0);
        sleep(5*60);
        return;
      }

      if(wait<1) 
	/* Force sending NTP and waiting 1s */
	wait = ntp_next_signal_time = 
          ntp_last_request.time.tv_sec = 1;

      if(read_ntp(wait)>=0) {
        ntp_error_count=0;
        if( now.time.tv_sec < req.time.tv_sec + 15 ) continue;
	if(now.time.tv_sec > goal - 100 ||
           ABS(pts[2].server_minus_host) > ABS(pts[1].server_minus_host)+0.005+
           pts[2].host_time.esterror * 1e-6) {
          dpr(8,"Found suitable time point%s","");
          pts[0]=pts[1];
          pts[1]=pts[2];
          break;
        }
      }
      else if(ntp_error_count++ > 5) {
        send_warning("More than 5 invalid NTP packets. Waiting");
        sleep ( ntp_error_count += ntp_error_count/2 );
      }
    }
  }
}
    
    
int main(void)
{
  prog_start.modes=0;
  adjtimex(&prog_start);
  dpr(9,"Starting ntpwatch%s","\n");
  ntp_server.sin_family = AF_INET;
  ntp_server.sin_port   = htons(123);
  ntp_server.sin_addr.s_addr = 0;
  for(;;) {
    initialize_ntpwatch();
    fix_time();
  }
  return(0);
}
    
      

